package iHMMuneAlign;

import java.io.PrintStream;
import java.io.PrintWriter;
/**
 * this class is a container for one or more static methods used to
 * calculate the A value for a given VGene
 * The A-value for an alignment is the probability of an alignment in the VGene
 * at pos 0, this probability decays exponentially as we move away from
 * the start of the VGene
 */

public class A_Score
{
    	/**
     * the A value used for the CDR prediction is found
     * by looking at the number of mutations in the common
     * region of a sequence over the number of nucleotides in
     * this common region multiplied by the A_slope for this
     * common region.
     *
     * The A_slope is previously determined by using linear regression
     * on the A values (Y's) over the mutation ratio for the common region (X's)
     * and thus estimating the relationship between A value and number of 
     * of mutations for common region
     */

    static final char N_NUCLEOTIDE = 110;
    static final char X_NUCLEOTIDE = 120;
    static final int COMMON_AREA_START_POS = 129;
    static final int COMMON_AREA_END_POS = 243;
    static final int COMMON_AREA_NUCL_LENGTH = 115;
    static final double A_SLOPE = 0.0064999999999999997D;
    static final double B_ADDITION = 0.0015D;
    	
     /**
     * calculate the A probability using the above rules and
     * the number of mutations found in the V-Gene "VGene" when
     * compared to it's mathcing sequence "matching sequence", the
     * "VGene_start_offset" must be taken into account when placing
     * the common region
     * 
     * Pre: 	"VGene" and "matching_sequence" are aligned
     *
     * returns		A probability based on number of mutations in VGene's common area plus above rules
     * 
     * Errors	if VGene does not contain the entire common region, and
     *			if matching sequence is not at least the length of the VGene
     */

    public A_Score()
    {
    }

    public static double A_probability(String VGeneUnknownCase, String matching_sequenceUnknownCase, int VGene_start_offset, PrintWriter pw)
    {
        String VGene = VGeneUnknownCase.toLowerCase();
        String matching_sequence = matching_sequenceUnknownCase.toLowerCase();
        System.out.println("VGene length = " + VGene.length());
        System.out.println("Matching seq. length = " + matching_sequenceUnknownCase.length());
        if(VGene.length() + VGene_start_offset < 243)
        {
            //throw new Error("A_Score: A_probability(): VGene does not contain common region (VGene too short)");
            return 2;
        }
        if(VGene_start_offset > 129)
        {
            //throw new Error("A_Score: A_probability(): VGene does not contain common region (VGene offset too great)");
            return 3;
        }
        if(VGene.length() > matching_sequence.length())
        {
            //throw new Error("A_Score: A_probability(): VGene length is greater than its matching sequence length");
            return 4;
        }
        int mutationsInVGene = mutationsInVGene(VGeneUnknownCase, matching_sequenceUnknownCase, VGene_start_offset, pw);
        int vGeneLength = VGeneUnknownCase.length();
        double CRtoVGeneLengthRatio = 115D / (double)vGeneLength;
        double averageMutationsInCR = (double)mutationsInVGene * CRtoVGeneLengthRatio;
        int number_of_mutations_in_CR = 0;
        int number_of_NandX_nts_in_CR = 0;
        for(int i = 128 - VGene_start_offset; i <= 242 - VGene_start_offset; i++)
        {
            char V_nucl = VGene.charAt(i);
            char MS_nucl = matching_sequence.charAt(i);
            if(V_nucl != MS_nucl)
            {
                if(MS_nucl == 'n' || MS_nucl == 'x')
                {
                    number_of_NandX_nts_in_CR++;
                } else
                {
                    number_of_mutations_in_CR++;
                }
            }
        }

        System.out.println("number of mutations found in common area = " + number_of_mutations_in_CR);
        System.out.println("number of n/x nts. found in common area = " + number_of_NandX_nts_in_CR);
        if(pw != null)
        {
            pw.println(number_of_mutations_in_CR + "\t" + 115);
        }
        double probabilityMutationPlusB = (double)number_of_mutations_in_CR * 0.0064999999999999997D + 0.0015D;
        double NandXaddition = (averageMutationsInCR / 115D) * 0.0064999999999999997D * (double)number_of_NandX_nts_in_CR;
        double aProbability = probabilityMutationPlusB + NandXaddition;
        return aProbability;
    }

    private static int mutationsInVGene(String VGeneUnknownCase, String matching_sequenceUnknownCase, int VGene_start_offset, PrintWriter pw)
    {
        String VGene = VGeneUnknownCase.toLowerCase();
        String matching_sequence = matching_sequenceUnknownCase.toLowerCase();
        int number_of_mutations_in_VGene = 0;
        int number_of_NandX_nts_in_VGene = 0;
        for(int i = 0; i < VGene.length(); i++)
        {
            char V_nucl = VGene.charAt(i);
            char MS_nucl = matching_sequence.charAt(i);
            if(V_nucl != MS_nucl)
            {
                if(MS_nucl == 'n' || MS_nucl == 'x')
                {
                    number_of_NandX_nts_in_VGene++;
                } else
                {
                    number_of_mutations_in_VGene++;
                }
            }
        }

        int vGeneLength = VGene.length();
        System.out.println("number of mutations found in VGene area = " + number_of_mutations_in_VGene);
        System.out.println("number of n/x nts. found in VGene  area = " + number_of_NandX_nts_in_VGene);
        System.out.println("VGene area length = " + vGeneLength);
        if(pw != null)
        {
            pw.print(number_of_mutations_in_VGene + "\t" + VGene_start_offset + "\t" + vGeneLength + "\t");
        }
        return number_of_mutations_in_VGene;
    }

    public static void main(String args[])
    {
        double A_prob = A_probability("aaacccgggtttaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" +
"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" +
"aaaaaagaatacaa"
, "aaacccaaatttaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" +
"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" +
"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaggggggggggggggggggggggggggggggg"
, 128, null);
        System.out.println("A_prob = " + A_prob);
    }
}